@@ -4,6 +4,7 @@ from __future__ import annotations
 
 import itertools
 import json
+import math
 import re
 from collections import defaultdict
 from collections.abc import Callable, Iterable, Mapping
@@ -11,10 +12,11 @@ from fnmatch import fnmatchcase
 from glob import escape as escape_glob
 
 import awkward as ak
+from awkward._backends.dispatch import regularize_backend
 from awkward._backends.numpy import NumpyBackend
 from awkward._meta.meta import Meta
 from awkward._nplikes.numpy_like import NumpyMetadata
-from awkward._nplikes.shape import ShapeItem, unknown_length
+from awkward._nplikes.shape import unknown_length
 from awkward._parameters import (
     parameters_are_equal,
     parameters_union,
@@ -29,7 +31,6 @@ from awkward._typing import (
     Self,
     Tuple,
     TypeAlias,
-    Union,
 )
 
 __all__ = ("Form", "from_dict", "from_json", "from_type", "reserved_nominal_parameters")
@@ -37,7 +38,7 @@ __all__ = ("Form", "from_dict", "from_json", "from_type", "reserved_nominal_para
 np = NumpyMetadata.instance()
 numpy_backend = NumpyBackend.instance()
 
-FormKeyPathT: TypeAlias = Tuple[Union[str, int, None], ...]
+FormKeyPathT: TypeAlias = Tuple[str | int | None, ...]
 
 reserved_nominal_parameters: Final = frozenset(
     {
@@ -193,8 +194,8 @@ def from_dict(input: Mapping) -> Form:
             form_key=form_key,
         )
 
-    elif input["class"] == "VirtualArray":
-        raise ValueError("Awkward 1.x VirtualArrays are not supported")
+    elif input["class"] == "VirtualNDArray":
+        raise ValueError("Awkward 1.x VirtualNDArrays are not supported")
 
     else:
         raise ValueError(
@@ -288,7 +289,7 @@ def _expand_braces(text, seen=None):
     else:
         for combo in itertools.product(*alts):
             replaced = list(text)
-            for (start, stop), replacement in zip(spans, combo):
+            for (start, stop), replacement in zip(spans, combo, strict=True):
                 replaced[start:stop] = replacement
             yield from _expand_braces("".join(replaced), seen)
 
@@ -373,11 +374,11 @@ def regularize_buffer_key(buffer_key: str | Callable) -> Callable[[Form, str], s
 
 
 index_to_dtype: Final[dict[str, DType]] = {
-    "i8": np.dtype("<i1"),
-    "u8": np.dtype("<u1"),
-    "i32": np.dtype("<i4"),
-    "u32": np.dtype("<u4"),
-    "i64": np.dtype("<i8"),
+    "i8": np.dtype("i1"),
+    "u8": np.dtype("u1"),
+    "i32": np.dtype("i4"),
+    "u32": np.dtype("u4"),
+    "i64": np.dtype("i8"),
 }
 
 
@@ -528,6 +529,10 @@ class Form(Meta):
                 "The `highlevel=True` variant of `Form.length_zero_array` has been removed. "
                 "Please use `ak.Array(form.length_zero_array(...), behavior=...)` if an `ak.Array` is required.",
             )
+        if not regularize_backend(backend).nplike.known_data:
+            return self.length_zero_array(
+                backend=numpy_backend, highlevel=highlevel, behavior=behavior
+            ).to_typetracer()
 
         return ak.operations.ak_from_buffers._impl(
             form=self,
@@ -540,6 +545,7 @@ class Form(Meta):
             behavior=behavior,
             attrs=None,
             simplify=False,
+            enable_virtualarray_caching=True,
         )
 
     def length_one_array(
@@ -550,136 +556,134 @@ class Form(Meta):
                 "The `highlevel=True` variant of `Form.length_one_array` has been removed. "
                 "Please use `ak.Array(form.length_one_array(...), behavior=...)` if an `ak.Array` is required.",
             )
-
-        # The naive implementation of a length-1 array requires that we have a sufficiently
-        # large buffer to be able to build _any_ subtree.
-        def max_prefer_unknown(this: ShapeItem, that: ShapeItem) -> ShapeItem:
-            if this is unknown_length:
-                return this
-            if that is unknown_length:
-                return that
-            return max(this, that)
+        if not regularize_backend(backend).nplike.known_data:
+            return self.length_one_array(
+                backend=numpy_backend, highlevel=highlevel, behavior=behavior
+            ).to_typetracer()
 
         container = {}
 
-        def prepare_empty(form):
+        def prepare(form, length):
             form_key = f"node-{len(container)}"
 
-            if isinstance(form, (ak.forms.BitMaskedForm, ak.forms.ByteMaskedForm)):
-                container[form_key] = b""
-                return form.copy(content=prepare_empty(form.content), form_key=form_key)
-
-            elif isinstance(form, ak.forms.IndexedOptionForm):
-                container[form_key] = b""
-                return form.copy(content=prepare_empty(form.content), form_key=form_key)
-
-            elif isinstance(form, ak.forms.EmptyForm):
+            if isinstance(form, ak.forms.EmptyForm):
+                if length != 0:
+                    raise TypeError(
+                        "cannot generate a length_one_array from a Form with an "
+                        "unknowntype that cannot be hidden (EmptyForm requires length=0)"
+                    )
                 return form
 
-            elif isinstance(form, ak.forms.UnmaskedForm):
-                return form.copy(content=prepare_empty(form.content))
-
-            elif isinstance(form, (ak.forms.IndexedForm, ak.forms.ListForm)):
-                container[form_key] = b""
-                return form.copy(content=prepare_empty(form.content), form_key=form_key)
-
-            elif isinstance(form, ak.forms.ListOffsetForm):
-                container[form_key] = b""
-                return form.copy(content=prepare_empty(form.content), form_key=form_key)
-
-            elif isinstance(form, ak.forms.RegularForm):
-                return form.copy(content=prepare_empty(form.content))
-
             elif isinstance(form, ak.forms.NumpyForm):
-                container[form_key] = b""
+                dtype = ak.types.numpytype.primitive_to_dtype(form._primitive)
+                size = length * dtype.itemsize
+                for x in form.inner_shape:
+                    if x is not unknown_length:
+                        size *= x
+                container[form_key] = b"\x00" * size
                 return form.copy(form_key=form_key)
 
-            elif isinstance(form, ak.forms.RecordForm):
-                return form.copy(contents=[prepare_empty(x) for x in form.contents])
+            elif isinstance(form, ak.forms.UnmaskedForm):
+                # no buffer, content length = length
+                return form.copy(content=prepare(form.content, length))
 
-            elif isinstance(form, ak.forms.UnionForm):
-                # both tags and index will get this buffer
-                container[form_key] = b""
+            elif isinstance(form, ak.forms.BitMaskedForm):
+                # mask buffer with values that indicate None
+                dtype = index_to_dtype[form.mask]
+                mask_length = math.ceil(length / 8.0)
+                if form.valid_when:
+                    container[form_key] = b"\x00" * (mask_length * dtype.itemsize)
+                else:
+                    container[form_key] = b"\xff" * (mask_length * dtype.itemsize)
+                # content length must equal length
+                next_length = length
                 return form.copy(
-                    contents=[prepare_empty(x) for x in form.contents],
-                    form_key=form_key,
+                    content=prepare(form.content, next_length), form_key=form_key
                 )
 
-            else:
-                raise AssertionError(f"not a Form: {form!r}")
-
-        def prepare(form, multiplier):
-            form_key = f"node-{len(container)}"
-
-            if isinstance(form, (ak.forms.BitMaskedForm, ak.forms.ByteMaskedForm)):
+            elif isinstance(form, ak.forms.ByteMaskedForm):
+                # mask buffer with values that indicate None
+                dtype = index_to_dtype[form.mask]
                 if form.valid_when:
-                    container[form_key] = b"\x00" * multiplier
+                    container[form_key] = b"\x00" * (length * dtype.itemsize)
                 else:
-                    container[form_key] = b"\xff" * multiplier
-                # switch from recursing down `prepare` to `prepare_empty`
-                return form.copy(content=prepare_empty(form.content), form_key=form_key)
+                    container[form_key] = b"\x01" * (length * dtype.itemsize)
+                # content length must equal mask length
+                next_length = length
+                return form.copy(
+                    content=prepare(form.content, next_length), form_key=form_key
+                )
 
             elif isinstance(form, ak.forms.IndexedOptionForm):
-                container[form_key] = b"\xff\xff\xff\xff\xff\xff\xff\xff"  # -1
-                # switch from recursing down `prepare` to `prepare_empty`
-                return form.copy(content=prepare_empty(form.content), form_key=form_key)
-
-            elif isinstance(form, ak.forms.EmptyForm):
-                # no error if protected by non-recursing node type
-                raise TypeError(
-                    "cannot generate a length_one_array from a Form with an "
-                    "unknowntype that cannot be hidden (EmptyForm not within "
-                    "BitMaskedForm, ByteMaskedForm, or IndexedOptionForm)"
+                # index buffer with all -1
+                dtype = index_to_dtype[form.index]
+                minus_one = b"\xff" * dtype.itemsize
+                container[form_key] = minus_one * length
+                # with all -1 indices, next_length = max(0, max(index) + 1) = 0
+                next_length = 0
+                return form.copy(
+                    content=prepare(form.content, next_length), form_key=form_key
                 )
 
-            elif isinstance(form, ak.forms.UnmaskedForm):
-                return form.copy(content=prepare(form.content, multiplier))
-
-            elif isinstance(form, (ak.forms.IndexedForm, ak.forms.ListForm)):
-                container[form_key] = b"\x00" * (8 * multiplier)
+            elif isinstance(form, ak.forms.IndexedForm):
+                # index buffer with all zeros
+                dtype = index_to_dtype[form.index]
+                container[form_key] = b"\x00" * (length * dtype.itemsize)
+                # with all-zero indices, next_length = max(index) + 1 = 1 (if length > 0)
+                next_length = 1 if length > 0 else 0
                 return form.copy(
-                    content=prepare(form.content, multiplier), form_key=form_key
+                    content=prepare(form.content, next_length), form_key=form_key
+                )
+
+            elif isinstance(form, ak.forms.ListForm):
+                # starts and stops buffers, both all zeros (share same buffer via form_key)
+                # use max itemsize in case starts and stops have different dtypes
+                itemsize = max(
+                    index_to_dtype[form.starts].itemsize,
+                    index_to_dtype[form.stops].itemsize,
+                )
+                container[form_key] = b"\x00" * (length * itemsize)
+                # with starts=stops=0, next_length = max(stops[starts != stops]) = 0
+                next_length = 0
+                return form.copy(
+                    content=prepare(form.content, next_length), form_key=form_key
                 )
 
             elif isinstance(form, ak.forms.ListOffsetForm):
-                # offsets length == array length + 1
-                container[form_key] = b"\x00" * (8 * (multiplier + 1))
+                # offsets buffer has length+1 elements, all zeros
+                dtype = index_to_dtype[form.offsets]
+                container[form_key] = b"\x00" * ((length + 1) * dtype.itemsize)
+                # with all-zero offsets, next_length = offsets[-1] = 0
+                next_length = 0
                 return form.copy(
-                    content=prepare(form.content, multiplier), form_key=form_key
+                    content=prepare(form.content, next_length), form_key=form_key
                 )
 
             elif isinstance(form, ak.forms.RegularForm):
                 size = form.size
-
                 # https://github.com/scikit-hep/awkward/pull/2499#discussion_r1220503454
                 if size is unknown_length:
                     size = 1
-
-                return form.copy(content=prepare(form.content, multiplier * size))
-
-            elif isinstance(form, ak.forms.NumpyForm):
-                dtype = ak.types.numpytype.primitive_to_dtype(form._primitive)
-                size = multiplier * dtype.itemsize
-                for x in form.inner_shape:
-                    if x is not unknown_length:
-                        size *= x
-
-                container[form_key] = b"\x00" * size
-                return form.copy(form_key=form_key)
+                next_length = length * size
+                return form.copy(content=prepare(form.content, next_length))
 
             elif isinstance(form, ak.forms.RecordForm):
-                return form.copy(
-                    # recurse down all contents
-                    contents=[prepare(x, multiplier) for x in form.contents]
-                )
+                # no buffer, all contents have same length
+                return form.copy(contents=[prepare(x, length) for x in form.contents])
 
             elif isinstance(form, ak.forms.UnionForm):
-                # both tags and index will get this buffer, but index is 8 bytes
-                container[form_key] = b"\x00" * (8 * multiplier)
-                # recurse down contents[0] with `prepare`, but others with `prepare_empty`
-                contents = [prepare(form.contents[0], multiplier)]
-                for x in form.contents[1:]:
-                    contents.append(prepare_empty(x))
+                # tags all 0, index all 0
+                dtype_index = index_to_dtype[form.index]
+                container[form_key] = b"\x00" * (length * dtype_index.itemsize)
+                # content[0]: next_length = max(index[tags==0]) + 1 = 1 (if length > 0)
+                # content[i>0]: next_length = 0 (no elements with those tags)
+                contents = []
+                for i, content_form in enumerate(form.contents):
+                    if i == 0:
+                        next_length = 1 if length > 0 else 0
+                    else:
+                        next_length = 0
+                    contents.append(prepare(content_form, next_length))
                 return form.copy(contents=contents, form_key=form_key)
 
             else:
@@ -696,6 +700,7 @@ class Form(Meta):
             behavior=behavior,
             attrs=None,
             simplify=False,
+            enable_virtualarray_caching=True,
         )
 
     def _expected_from_buffers(
@@ -744,3 +749,35 @@ class Form(Meta):
             and not (form_key and self._form_key != other._form_key)
             and compare_parameters(self._parameters, other._parameters)
         )
+
+
+def form_with_unique_keys(form: Form, key: tuple) -> Form:
+    def impl(form: Form, key: tuple) -> None:
+        # Set form key as str
+        # Can be interpreted as tuple again with `ast.literal_eval(form_key)`
+        form.form_key = repr(key)
+
+        # If the form is a record we need to loop over all fields in the
+        # record and set form that include the field name; this will keep
+        # recursing as well.
+        if form.is_record:
+            for field in form.fields:
+                impl(form.content(field), (*key, field))  # type: ignore[attr-defined]
+
+        elif form.is_union:
+            for i, entry in enumerate(form.contents):  # type: ignore[attr-defined]
+                impl(entry, (*key, i))
+
+        # NumPy like array is easy
+        elif form.is_numpy or form.is_unknown:
+            pass
+
+        # Anything else grab the content and keep recursing
+        else:
+            # use `None` for anything else, could be more informative at some point, but there's no need for it right now
+            impl(form.content, (*key, None))  # type: ignore[attr-defined]
+
+    # Perform a "deep" copy without preserving references
+    form = ak.forms.from_dict(form.to_dict())
+    impl(form, key)
+    return form
